from . import const, OppleDevice

MESSAGE_TYPE     = const.MESSAGE_TYPE
QUERY_RES_OFFSET = const.QUERY_RES_OFFSET


class OppleLightDevice(OppleDevice.OppleDevice):
    def __init__(self, ip='', message=None):
        self._isPowerOn = False
        self._brightness = 0
        self._colorTemperature = 0
        super().__init__(ip, message)

    def init(self, message):
        super(OppleLightDevice, self).init(message)
        self.update()

    def update(self):
        if not self.is_init:
            self.async_init()

        if not self.is_init:
            return

        message = self.send('QUERY', reply=True)
        if message:
            self._isPowerOn = message.get(QUERY_RES_OFFSET['POWER_ON'], 1, int)
            self._brightness = message.get(QUERY_RES_OFFSET['BRIGHT'], 1, int)
            self._colorTemperature = message.get(QUERY_RES_OFFSET['COLOR_TEMP'], 2, int)

    def set(self, message_type, value, check=None, _time=3):
        if _time == 0:
            return False

        if check():
            return True

        self.send(message_type, value)

        if check:
            self.update()
            if not check():
                return self.set(message_type, value, check, _time-1)

        return True

    @property
    def power_on(self):
        return self._isPowerOn

    @power_on.setter
    def power_on(self, value):
        value = 1 if value else 0

        def check():
            return self.power_on == (value == 1)

        self.set('POWER_ON', value.to_bytes(1, 'big'), check)

    @property
    def brightness(self):
        return self._brightness

    @brightness.setter
    def brightness(self, value):
        value = max(0, value)
        value = min(255, value)

        def check():
            return self.brightness == value

        self.set('BRIGHTNESS', value.to_bytes(1, 'big'), check)

    @property
    def color_temperature(self):
        return self._colorTemperature

    @color_temperature.setter
    def color_temperature(self, value):
        value = max(2700, value)
        value = min(6500, value)

        def check():
            return self.color_temperature == value

        self.set('COLOR_TEMP', value.to_bytes(2, 'big'), check)



import socket

from . import const, Message

MESSAGE_TYPE = const.MESSAGE_TYPE
SEARCH_RES_OFFSET = const.SEARCH_RES_OFFSET


class OppleDevice(object):
    def __init__(self, ip='', message=None):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.bind(('0.0.0.0', 0))
        self.socket.settimeout(2)

        self.id = 0
        self.port = 0
        self.version = 0
        self.name = ''
        self.ip = ip
        self.ip_raw = ()
        self.mac = ''
        self.mac_raw = ()
        self.server_port = self.socket.getsockname()[1]
        self.is_init = False
        self.is_online = False

        if message:
            self.init(message)
        else:
            self.async_init()

    def init(self, message):
        self.id = message.get(SEARCH_RES_OFFSET['ID_LOW'], 4, value_type=int)
        self.port = message.get(SEARCH_RES_OFFSET['PORT'], 2, value_type=int)
        self.version = message.get(SEARCH_RES_OFFSET['VERSION'], 4, value_type=int)
        self.name = str(message.get(SEARCH_RES_OFFSET['NAME'], 0xE), 'gbk')
        self.ip_raw = message.get(SEARCH_RES_OFFSET['IP'], 4)
        self.mac_raw = message.get(SEARCH_RES_OFFSET['MAC'], 6)

        self.ip = '.'.join(map(lambda x: str(x), self.ip_raw))
        self.mac = ':'.join(map(lambda x: hex(x)[2:], self.mac_raw))

        self.is_init = True
        self.is_online = True

    def async_init(self):
        self.port = const.BROADCAST_PORT
        message = self.send('SEARCH', reply=True)
        if message:
            self.init(message)

    def send(self, message_type, data=None, reply=False):
        message = Message.build_message(MESSAGE_TYPE[message_type], data, self)
        self.socket.sendto(message.data, (self.ip, self.port))

        if reply:
            while True:
                try:
                    data, address = self.socket.recvfrom(1024)
                    incoming_message = Message.parse_message(data, self)

                    if message.get_request_sn() != incoming_message.get_response_sn():
                        continue

                    self.is_online = True
                    return incoming_message

                except socket.timeout:
                    self.is_online = False
                    return None


def search():
    message = Message.build_message(MESSAGE_TYPE['SEARCH'])
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    s.sendto(message.to_bytes(), ('255.255.255.255', const.BROADCAST_PORT))
    s.settimeout(2)

    from . import OppleLightDevice

    device_list = []

    while True:
        try:
            data, address = s.recvfrom(1024)
            message = Message.parse_message(data)

            if message.get(SEARCH_RES_OFFSET['CLASS_SKU'], 4, value_type=int) == 0x100010E:
                device = OppleLightDevice.OppleLightDevice(message=message)
            else:
                break

            device_list.append(device)
            yield device
        except socket.timeout:
            break

    return device_list



import random
import crc16
from . import const

MESSAGE_OFFSET = const.MESSAGE_OFFSET


class Message(object):

    def __init__(self, device, data=bytearray()):
        self.device = device
        self.data = bytearray(data)
        self.body = data[MESSAGE_OFFSET['BODY']:]

    def get(self, offset, length=4, value_type=None, header=False):
        if not header:
            offset += MESSAGE_OFFSET['BODY']

        value = self.data[offset : offset + length]

        if value_type == int:
            return int.from_bytes(value, 'big')
        else:
            return value

    def set(self, value, offset, length=4, header=False):
        if not header:
            offset += MESSAGE_OFFSET['BODY']

        self.data[offset : offset + length] = value.to_bytes(length, 'big')

    def set_checksum(self):
        crc = crc16.crc16xmodem(bytes(self.data[0x64:]))
        self.set(crc, MESSAGE_OFFSET['CHECK_SUM'], header=True)

    def encrypt(self):
        if self.device is None or not self.device.is_init:
            return

        password = bytearray(20)
        mac = self.device.mac_raw

        for i in range(0, 20):
            password[i] = mac[i % mac.__len__()]

        for i, v in enumerate(self.body):
            self.data[i + MESSAGE_OFFSET['BODY']] = v ^ password[i % password.__len__()]

    def decrypt(self):
        return self.encrypt()

    def get_request_sn(self):
        return self.data[MESSAGE_OFFSET['REQ_SERIAL_NUM'] : MESSAGE_OFFSET['REQ_SERIAL_NUM'] + 4]

    def get_response_sn(self):
        return self.data[MESSAGE_OFFSET['RES_SERIAL_NUM'] : MESSAGE_OFFSET['RES_SERIAL_NUM'] + 4]

    def to_bytes(self):
        return bytes(self.data)


def build_message(msg_type, body=None, device=None):
    sn = random.randint(1, 1000)
    body = bytearray() if body is None else bytearray(body)
    data = bytearray(MESSAGE_OFFSET['BODY']) + body

    message = Message(device, data)

    message.set(0x03F2, MESSAGE_OFFSET['L2_TYPE'], header=True)
    message.set(0x2775, MESSAGE_OFFSET['L3_VERSION'], header=True)
    message.set(0x0001, MESSAGE_OFFSET['L3_ID'], header=True)
    message.set(0x0002, MESSAGE_OFFSET['OFFSET'], header=True)
    message.set(0x0003, MESSAGE_OFFSET['TTL'], header=True)
    message.set(0x0005, MESSAGE_OFFSET['L3_CHECKSUM'], header=True)
    message.set(0x0001, MESSAGE_OFFSET['DEST_OBJ_TYPE'], header=True)
    message.set(0x6A68, MESSAGE_OFFSET['SRC_ID'], header=True)

    message.set(sn, MESSAGE_OFFSET['REQ_SERIAL_NUM'], header=True)
    message.set(msg_type, MESSAGE_OFFSET['MSG_TYPE'], header=True)
    message.set(body.__len__() + 0x68, MESSAGE_OFFSET['PKG_LENGTH'], header=True)
    message.set(body.__len__() + 0x18, MESSAGE_OFFSET['MSG_LENGTH'], header=True)

    if device and device.is_init:
        message.set(device.server_port, MESSAGE_OFFSET['DEST_PORT'], header=True)
        message.set(device.id, MESSAGE_OFFSET['DEST_ID'], header=True)

    message.set_checksum()
    message.encrypt()

    return message


def parse_message(data, device=None):
    message = Message(device, data)
    message.decrypt()
    return message



import argparse

from . import OppleDevice
from . import OppleLightDevice


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('action', help='"get" "set" or "search"')
    parser.add_argument('-a', '--address', help='device ip address', default='255.255.255.255')
    parser.add_argument('-p', '--power', help='light power status, "on" or "off"')
    parser.add_argument('-b', '--brightness', help='light brightness, 10 <= x <= 255')
    parser.add_argument('-c', '--ct', help='light color temperature, 2700 <= x <= 6500')

    args = parser.parse_args()

    if args.action == 'search':
        for device in OppleDevice.search():
            print("%s\t%s" % (device.ipstr, device.macstr))

    elif args.action == 'set':
        if args.address is None:
            return print('IP address is empty.')

        device = OppleLightDevice.OppleLightDevice(ip=args.address)

        if args.power is None and args.brightness is None and args.ct is None:
            return print('nothing to change.')

        if args.power is not None:
            if args.power != 'on' and args.power != 'off':
                return print('power can only be "on" or "off"')
            device.power_on = args.power == 'on'

        if args.brightness is not None:
            try:
                value = int(args.brightness)
            except (TypeError, ValueError):
                return print("brightness value is illegal")

            if value < 0 or value > 255:
                return print("brightness is out of range")
            device.brightness = value

        if args.ct is not None:
            try:
                value = int(args.ct)
            except (TypeError, ValueError):
                return print("color temperature value is illegal")

            if value < 2700 or value > 6500:
                return print("color temperature is out of range")
            device.color_temperature = value

        print("Light %s" % device.ip)
        if args.power is not None:
            print("Power:\t\t\t%s" % "on" if device.power_on else "off")
        if args.brightness is not None:
            print("Brightness:\t\t%s" % device.brightness)
        if args.ct is not None:
            print("Color Temperature:\t%s" % device.color_temperature)
    else:
        if args.address is None:
            print('IP address is empty.')
            exit(0)

        device = OppleLightDevice.OppleLightDevice(ip=args.address)

        if device.is_online:
            print("Light %s\nPower:\t\t\t%s\nBrightness:\t\t%s\nColor Temperature:\t%s"
                  % (device.ip, "on" if device.power_on else "off", device.brightness, device.color_temperature))
        else:
            print("Light %s is offline." % device.ip)


if __name__ == "__main__":
    main()
    exit(0)



from setuptools import setup, find_packages


setup(
    name='pyoppleio',
    version='1.0.0',
    keywords=('opple', 'iot'),
    description='Python library for interfacing with opple mobile control light',
    long_description=open('README.rst', 'rt').read(),
    author='jedmeng',
    author_email='jedm@jedm.cn',
    url='https://github.com/jedmeng/python-oppleio',
    license='MIT',
    install_requires=[
        'crc16==0.1.1'
    ],
    packages=find_packages(),
    platforms='any',
    python_requires='>=3.5',
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'oppleio=pyoppleio.__main__:main',
        ]
    },
)



BROADCAST_PORT = 0xD6D9

MESSAGE_TYPE = {
    'POWER_ON'  : 0x3110000,
    'BRIGHTNESS': 0x3130000,
    'COLOR_TEMP': 0x31b0000,
    'QUERY'     : 0x30f0000,
    'SEARCH'    : 0x2010000,
}

MESSAGE_OFFSET = {
    'SRC_IP'        : 0x00,
    'SRC_PORT'      : 0x04,
    'DEST_IP'       : 0x08,
    'DEST_PORT'     : 0x0C,
    'L2_TYPE'       : 0x10,
    'L3_VERSION'    : 0x14,
    'SERVER_TYPE'   : 0x18,
    'PKG_LENGTH'    : 0x1C,
    'L3_ID'         : 0x20,
    'OFFSET'        : 0x24,
    'TTL'           : 0x28,
    'PROP'          : 0x2C,
    'L3_CHECKSUM'   : 0x30,
    'SRC_ADD_TYPE'  : 0x34,
    'SRC_OBJ_TYPE'  : 0x38,
    'SRC_ID'        : 0x3C,
    'SRC_ID_2'      : 0x40,
    'SRC_ID_3'      : 0x44,
    'SRC_ID_4'      : 0x48,
    'DEST_ADD_TYPE' : 0x4C,
    'DEST_OBJ_TYPE' : 0x50,
    'DEST_ID'       : 0x54,
    'DEST_ID_2'     : 0x58,
    'DEST_ID_3'     : 0x5C,
    'DEST_ID_4'     : 0x60,
    'REQ_SERIAL_NUM': 0x64,
    'RES_SERIAL_NUM': 0x68,
    'MSG_LENGTH'    : 0x6C,
    'CHECK_SUM'     : 0x70,
    'MSG_TYPE'      : 0x74,
    'RESERVE'       : 0x78,
    'BODY'          : 0x7C
}

SEARCH_RES_OFFSET = {
    'CLASS_SKU': 0x03,  # L: 0x4
    'SRC_TYPE' : 0x07,  # L: 0x2
    'MAC'      : 0x09,  # L: 0x8
    'OBJ_TYPE' : 0x11,  # L: 0x2
    'ID_LOW'   : 0x13,  # L: 0x4
    'ID_HIGH'  : 0x17,  # L: 0x4
    'VERSION'  : 0x1B,  # L: 0x4
    'IP'       : 0x1F,  # L: 0x4
    'PORT'     : 0x23,  # L: 0x2
    'NAME'     : 0x25,  # L: 0xe
    'IS_CLEAN' : 0x33,  # L: 0x1
}

QUERY_RES_OFFSET = {
    'POWER_ON'  : 0x1,  # L: 0x1
    'BRIGHT'    : 0x2,  # L: 0x1
    'COLOR_TEMP': 0x7,  # L: 0x2
}
